-- indexes to speed up queries checking if grants are still available
drop index if exists grants_checkin_time_idx;
drop index if exists grants_feature_id_idx;
drop index if exists checked_features_id_idx;
drop index if exists session_features_cluster_user_instance_time_idx;

create index grants_checkin_time_idx on i3dh.grants using btree (checkin_time);
create index grants_feature_id_idx on i3dh.grants using btree (feature_id);
create index checked_features_id_idx on i3dh.checked_features using btree (id);

-- index for querying time ranges more quickly
create index session_features_time_idx on i3dh.session_features using btree (time);
create index session_features_user_id_idx on i3dh.session_features using btree (user_id);
create index session_features_cluster_id_idx on i3dh.session_features using btree (cluster_id);
create index session_features_app_instance_id_idx on i3dh.session_features using btree (app_instance_id);
create index session_features_cluster_user_instance_time_idx on i3dh.session_features using btree (cluster_id asc, user_id asc, app_instance_id asc, "time" asc);

create or replace function i3dh.update_summary(p_features bigint[], p_costs bigint[], p_token_duration interval,
                                               p_idle_duration interval)
    returns void
as
$body$
declare
    first_it                bool = true;
    cluster_id_current      varchar;
    cluster_id_last         varchar;
    user_id_current         varchar;
    user_id_last            varchar;
    app_instance_id_current varchar;
    app_instance_id_last    varchar;
    features_current        int8;
    features_accumulated    int8 = 0;
    features_session        int8 = 0;
    time_start_session      timestamptz;
    time_start_interval     timestamptz;
    time_current            timestamptz;
    last_ping               timestamptz;
    costs                   int8 = 0;
begin
    truncate i3dh.summary restart identity;

    for cluster_id_current, user_id_current, app_instance_id_current, features_current, time_current in select sf.cluster_id,
                                                                                                               sf.user_id,
                                                                                                               sf.app_instance_id,
                                                                                                               sf.feature_mask,
                                                                                                               sf."time"
                                                                                                        from i3dh.session_features sf
                                                                                                        order by sf.cluster_id asc,
                                                                                                                 sf.user_id asc,
                                                                                                                 sf.app_instance_id asc,
                                                                                                                 sf."time" asc
        loop
            if first_it then
                time_start_session = time_current;
                time_start_interval = time_current;
                last_ping = time_current;
                cluster_id_last = cluster_id_current;
                user_id_last = user_id_current;
                app_instance_id_last = app_instance_id_current;

                first_it = false;
            end if;

            if ((cluster_id_current != cluster_id_last)
                or (user_id_current != user_id_last)
                or (app_instance_id_current != app_instance_id_last)
                or (time_current - last_ping > p_idle_duration))
            then
                costs = costs + i3dh.feature_costs(features_accumulated, p_features, p_costs);
                features_session = features_session | features_accumulated;

                --- add new entry to result set
                insert into i3dh.summary(cluster_id, user_id, app_instance_id, features, costs, start_time, duration)
                values (cluster_id_last, user_id_last, app_instance_id_last, features_session, costs,
                        time_start_session, last_ping - time_start_session);

                time_start_session = time_current;
                time_start_interval = time_current;
                last_ping = time_current;
                cluster_id_last = cluster_id_current;
                user_id_last = user_id_current;
                app_instance_id_last = app_instance_id_current;
                features_accumulated = features_current;
                features_session = 0;
                costs = 0;
            elsif (time_current - time_start_interval > p_token_duration)
            then
                features_accumulated = features_accumulated | features_current;
                costs = costs + i3dh.feature_costs(features_accumulated, p_features, p_costs);
                features_session = features_session | features_accumulated;

                time_start_interval = time_start_interval + p_token_duration;
                last_ping = time_start_interval + p_token_duration;
                cluster_id_last = cluster_id_current;
                user_id_last = user_id_current;
                app_instance_id_last = app_instance_id_current;
                features_accumulated = features_current;
            else
                --- do accumulation
                last_ping = time_current;
                features_accumulated = features_accumulated | features_current;
            end if;
        end loop;

    --- add last entry to result set
    if (cluster_id_last is not null) then
        costs = costs + i3dh.feature_costs(features_accumulated, p_features, p_costs);
        features_session = features_session | features_accumulated;

        insert into i3dh.summary(cluster_id, user_id, app_instance_id, features, costs, start_time, duration)
        values (cluster_id_last, user_id_last, app_instance_id_last, features_session, costs, time_start_session,
                last_ping - time_start_session);
    end if;

    refresh materialized view i3dh.summary_view;

    return;
end
$body$
    language 'plpgsql';

alter function i3dh.update_summary(bigint[], bigint[], interval, interval) owner to i3dhub;

create or replace function i3dh.feature_costs(p_feature bigint, p_features bigint[], p_costs bigint[])
    returns int8
as
$body$
declare
    r_sum bigint = 0;
begin
    for i in 0..array_length(p_features, 1)
        loop
            if (p_feature & p_features[i]) > 0 then
                r_sum = r_sum + p_costs[i];
            end if;
        end loop;

    return r_sum;
end
$body$
    language 'plpgsql';

alter function i3dh.feature_costs(bigint, bigint[], bigint[]) owner to i3dhub;

create or replace function i3dh.get_usage_summary(p_intervall varchar)
    returns table
            (
                r_date   timestamptz,
                r_tokens int8
            )
as
$body$
begin
    return query (
        select date_trunc(p_intervall, s.start_time), sum(s.costs)::int8
        from i3dh.summary_view s
        group by date_trunc(p_intervall, s.start_time)
        order by date_trunc(p_intervall, s.start_time) asc
    );
end
$body$
    language 'plpgsql';

alter function i3dh.get_usage_summary(varchar) owner to i3dhub;

create or replace function i3dh.get_usage_summary_today()
    returns table
            (
                r_date   timestamptz,
                r_tokens int8
            )
as
$body$
begin
    return query (
        select date_trunc('day', s.start_time), sum(s.costs)::int8
        from i3dh.summary_view s
        where date_trunc('day', now()) = date_trunc('day', s.start_time)
        group by date_trunc('day', s.start_time)
    );
end
$body$
    language 'plpgsql';

alter function i3dh.get_usage_summary_today() owner to i3dhub;

create or replace function i3dh.get_num_used_tokens()
    returns int8
as
$body$
begin
    return (
        select sum(s.costs)::int8
        from i3dh.summary_view s
    );
end
$body$
    language 'plpgsql';

alter function i3dh.get_num_used_tokens() owner to i3dhub;

create or replace function i3dh.num_active_sessions()
    returns int8
as
$body$
begin
    return (
        select count(*)
        from i3dh.summary_view s
        where now() - (s.start_time + s.duration) <= interval '5 minutes'
    );
end
$body$
    language 'plpgsql';

alter function i3dh.num_active_sessions() owner to i3dhub;
